from mininet.node import Switch
from mininet.node import Controller
from typing import Literal, NamedTuple
import pathlib
import os
import sys
import subprocess
Daemon = Literal["bgpd"]

def chown_recursive(dir: pathlib.Path, user: str, group: str):
    subprocess.run(["chown", "-R", f"{user}:{group}", str(dir)])
    
class FrrSwitch(Switch):
    daemons: list[Daemon]
    config_path: pathlib.Path
    
    def __init__(self, name, config_file: str | None = None, daemons: list[Daemon] = [], **params):
        if config_file == None:
            config_dir = pathlib.Path(sys.argv[0]).parent.joinpath("config")
            if os.path.exists(os.path.join(config_dir, name + ".cfg")):
                config_file = config_dir.joinpath(name + ".cfg")
            else:
                config_file = config_dir.joinpath(name + ".conf")
        else:
            config_file = pathlib.Path(config_file)
            
        if not config_file.exists():
            raise Exception(f"config path of {name} ({str(config_file)}) not exist!")
        
        # create a temp directory for each switch
        config_path = pathlib.Path("/tmp").joinpath("frrnet").joinpath(name)
        config_path.mkdir(parents=True, exist_ok=True)
        subprocess.run(["cp", str(config_file), str(config_path.joinpath("frr.conf"))])
        subprocess.run(["touch", str(config_path.joinpath("vtysh.conf"))])
        
        self.config_path = config_path
        chown_recursive(config_path, "frr", "frr")
        super().__init__(name, privateDirs=[("/etc/frr", str(config_path)), "/run/frr"], **params)
        self.daemons = daemons
        self.controlIntf = None
        
    def start(self, _controllers: list[Controller]):
        
        for daemon in ["zebra", "mgmtd", "staticd"]:
            self.cmd(f"/usr/libexec/frr/{daemon} -d")
            
        for daemon in self.daemons:
            self.cmd(f"/usr/libexec/frr/{daemon} -d")
    
        self.cmd("vtysh -b")
                
    def terminate(self):
        super().terminate()
        # remove the temp directory
        if self.config_path.exists():
            subprocess.run(["rm", "-rf", str(self.config_path)])
        
        
    
    def stop(self, deleteIntfs=True):
        for daemon in self.daemons:
            self.cmd(f"killall {daemon}")
            
        super().stop(deleteIntfs)



from typing import override
from mininet.cli import CLI
from mininet.log import error
import sys

class FrrCLI(CLI):
    helpStr = (
        'You may also send a network command to a node using:\n'
        '  <node> command {args}\n'
        'For example:\n'
        '  mininet> h1 show ip bgp\n'
        '\n'
        'If you want to send a shell command to node, use `sh` beforehands:\n'
        '  <node> bash command {args}\n'
        'For example:\n'
        '  mininet> h1 sh ifconfig\n'
        '\n'
        'The interpreter automatically substitutes IP addresses\n'
        'for node names when a node is the first arg, so commands\n'
        'like\n'
        '  mininet> h2 ping h3\n'
        'should work.\n'
        '\n'
        'Some character-oriented interactive commands require\n'
        'noecho:\n'
        '  mininet> noecho h2 vi foo.py\n'
        'However, starting up an xterm/gterm is generally better:\n'
        '  mininet> xterm h2\n\n'
    )
    def __init__(self, mininet, stdin=sys.stdin, script=None,
                 *args, **kwargs):
        super().__init__(mininet, stdin, script, *args, **kwargs)
        
    @override
    def default(self, line):
        first, args, line = self.parseline( line )
        
        if first in self.mn:
            if not args:
                error( '*** Please enter a command for node: %s <cmd>\n'
                       % first )
                return
            node = self.mn[ first ]
            rest = args.split( ' ' )
            
            if rest[0] != 'sh':
                rest = f"vtysh -c '{' '.join(rest)}'"
                node.sendCmd( rest )
                self.waitForNode( node )
                return
            else:
                rest = rest[1:]
                # Substitute IP addresses for node names in command
                # If updateIP() returns None, then use node name
                rest = [ self.mn[ arg ].defaultIntf().updateIP() or arg
                         if arg in self.mn else arg
                         for arg in rest ]
                rest_cmd = ' '.join(rest).replace("'", "\\'")
                rest = f"sh -c '{rest_cmd}'"
                node.sendCmd( rest )
                self.waitForNode( node )

        else:
            error( '*** Unknown command: %s\n' % line )


from mininet.net import Mininet
from mininet.node import Host
from mininet.link import Link, Intf
from frrnet.node import FrrSwitch


class FrrNet(Mininet):
    def __init__(self, topo=None, switch=FrrSwitch, host=Host, controller=None, link=Link, intf=Intf, build=True, xterms=False, cleanup=False, ipBase='10.0.0.0/8', inNamespace=True, autoSetMacs=True, autoStaticArp=False, autoPinCpus=False, listenPort=None, waitConnected=False):
        super().__init__(topo, switch, host, controller, link, intf, build, xterms, cleanup, ipBase, inNamespace, autoSetMacs, autoStaticArp, autoPinCpus, listenPort, waitConnected)
        
    def configureControlNetwork(self):
        pass


from mininet.topo import Topo
from mininet.log import setLogLevel
from frrnet.cli import FrrCLI
from frrnet.net import FrrNet
from typing import Type, TypeVar

T = TypeVar("T", bound=Topo)
def frrnet_main(TopoClass: Type[T]):
    setLogLevel("info")
    topo = TopoClass()
    
    net = FrrNet(topo)
    
    net.start()
    FrrCLI(net)
    net.stop()
    